#!/usr/bin/perl -s
#
# icmap+tops.pl - writes inverse mapping of CMap to PostScript file
#
# usage:
#	At first, please edit $in_dir and $out_dir.
#	$in_dir has ac13/, ag13/, aj14/, aj20/ and ak12/ directories.
#	$cfn_dir has CMaps for PDF Core fonts at http://partners.adobe.com/.
#	$uni_dir has Unicode mapping tables at ftp://ftp.unicode.org/.
#	In $out_dir, TrueType/* resources will be created.
#	% emacs icmap+tops.pl
#	% icmap+tops.pl
#
# Copyright (C) 2000/11/15-23,12/3-7,13,2001/1/4,3/30,31,4/2
# Taiji Yamada <taiji@aihara.co.jp>, gs-cjk project, All rights reserved.
# This software is provided under the BSD license. For full details,
# please see http://www.freebsd.org/copyright/license.html.
#
my $creator = 'Taiji Yamada <taiji@aihara.co.jp>';

#my $in_dir = '/usr/local/share/ghostscript/Resource';
#my $out_dir = 'TrueType-1999';
my $in_dir = '.';
my $out_dir = 'TrueType';

my $cfn_dir = 'CMap';
my $uni_dir = 'ftp.unicode.org/Public/MAPPINGS';

( ! -d $out_dir ) && mkdir $out_dir, 0755;
$macdos2 = "gawk 'BEGIN{RS=\"[\\r\\n][\\r\\n]?\"} {print}'";

my $cid2code_1999 = {
  'Adobe-CNS1'   => "$in_dir/ac10/cid2code.txt",
  'Adobe-GB1'    => "$in_dir/ag12/cid2code.txt",
  'Adobe-Japan1' => "$in_dir/aj12/cid2code.txt",
  'Adobe-Japan2' => "$in_dir/aj20/cid2code.txt",
  'Adobe-Korea1' => "$in_dir/ak11/cid2code.txt",
};
my $cid2code_2000 = {
  'Adobe-CNS1'   => "$in_dir/ac13/cid2code.txt",
  'Adobe-GB1'    => "$in_dir/ag13/cid2code.txt",
  'Adobe-Japan1' => "$in_dir/aj14/cid2code.txt",
  'Adobe-Japan2' => "$in_dir/aj20/cid2code.txt",
  'Adobe-Korea1' => "$in_dir/ak12/cid2code.txt",
};
#my $cid2code = $cid2code_1999;
my $cid2code = $cid2code_2000;

my $cid2ttf = {
  'Adobe,CNS1,Big5'       => { base=>'ETen-B5',      sup=>0, },
  'Adobe,CNS1,Unicode'    => { base=>'UniCNS-UCS2',  sup=>0, },
  'Adobe,GB1,PRC'         => { base=>'GBK-EUC',      sup=>1, },
  'Adobe,GB1,Unicode'     => { base=>'UniGB-UCS2',   sup=>1, },
  'Adobe,Japan1,ShiftJIS' => { base=>'90ms-RKSJ',    sup=>2, },
  'Adobe,Japan1,Unicode'  => { base=>'UniJIS-UCS2',  sup=>4, },
  'Adobe,Japan2,Unicode'  => { base=>'UniHojo-UCS2', sup=>0, },
  'Adobe,Korea1,Wansung'  => { base=>'KSCms-UHC',    sup=>1, },
  'Adobe,Korea1,Johab'    => { base=>'KSC-Johab',    sup=>1, },
  'Adobe,Korea1,Unicode'  => { base=>'UniKS-UCS2',   sup=>1, },
};

my $cmap_name = { # In each array, CMap names will be stored.
  'Adobe,CNS1'   => [],
  'Adobe,GB1'    => [],
  'Adobe,Japan1' => [],
  'Adobe,Japan2' => [],
  'Adobe,Korea1' => [],
};

my $itables = { # since 20010330, based on the idea of Hideyuki Suzuki
  'Adobe-CNS1-UCS2'   => { f=>"$cfn_dir/Adobe-CNS1-UCS2",   },
  'Adobe-GB1-UCS2'    => { f=>"$cfn_dir/Adobe-GB1-UCS2",    }, # ...
  'Adobe-Japan1-UCS2' => { f=>"$cfn_dir/Adobe-Japan1-UCS2", }, # ...
  'Adobe-Korea1-UCS2' => { f=>"$cfn_dir/Adobe-Korea1-UCS2", }, # ...
};
my $ctables = {
  'ETen-B5-UCS2'   => { f=>"$cfn_dir/ETen-B5-UCS2",    },
  'B5pc-UCS2'      => { f=>"$cfn_dir/B5pc-UCS2",       }, # usecmap
  'GBK-EUC-UCS2'   => { f=>"$cfn_dir/GBK-EUC-UCS2",    },
  'GBpc-EUC-UCS2'  => { f=>"$cfn_dir/GBpc-EUC-UCS2",   }, # usecmap ...
  '90ms-RKSJ-UCS2' => { f=>"$cfn_dir/90ms-RKSJ-UCS2",  },
  '90pv-RKSJ-UCS2' => { f=>"$cfn_dir/90pv-RKSJ-UCS2",  }, # usecmap ...
  'KSCms-UHC-UCS2' => { f=>"$cfn_dir/KSCms-UHC-UCS2",  },
 #'KSCpc-EUC-UCS2' => { f=>"$cfn_dir/KSCpc-EUC-UCS2",  }, # usecmap
  'KSCpc-EUC-UCS2' => { f=>"$cfn_dir/KSCpc-EUC-UCS2C", },
};
my $utables = {
  'ascii'        => {},
  'ms-cp950'     => { f=>"$uni_dir/VENDORS/MICSFT/WINDOWS/CP950.TXT", },
  'mac-big5'     => { f=>"$uni_dir/VENDORS/APPLE/CHINTRAD.TXT",       },
  'ms-cp936'     => { f=>"$uni_dir/VENDORS/MICSFT/WINDOWS/CP936.TXT", },
  'mac-gb'       => { f=>"$uni_dir/VENDORS/APPLE/CHINSIMP.TXT",       },
 #'mac-gbt'      => { f=>"$uni_dir/VENDORS/APPLE/CHINSIMP.TXT",       },
  'ms-cp932'     => { f=>"$uni_dir/VENDORS/MICSFT/WINDOWS/CP932.TXT", },
  'mac-japanese' => { f=>"$uni_dir/VENDORS/APPLE/JAPANESE.TXT",       },
  'ms-cp949'     => { f=>"$uni_dir/VENDORS/MICSFT/WINDOWS/CP949.TXT", },
  'mac-korean'   => { f=>"$uni_dir/VENDORS/APPLE/KOREAN.TXT",         },
 #'euc-kr'       => { f=>"$uni_dir/EASTASIA/KSC/KSC5601.TXT",         },
 #'johab'        => { f=>"$uni_dir/EASTASIA/KSC/JOHAB.TXT",           },
};
my $tables = undef;
$tables = {}					if ($GEN eq '0.81');
$tables = { %$utables }				if ($GEN eq '0.92');
$tables = { %$itables }				if ($GEN eq '0.925');
$tables = { %$ctables }				if ($GEN eq '0.93');
$tables = { %$itables, %$ctables }		if ($GEN eq '0.935' || !$GEN);
$tables = { %$itables, %$ctables, %$utables }	if ($GEN eq '0.94');

for my $table (keys %{$tables}) {
  $tables->{$table}->{map} = {};
  my $map = \$tables->{$table}->{map};
  if ($table =~ /^ascii/) {
    for (my $c=0x20; $c<=0x7e; $c++) {
      $$map->{$c} = $c;
    }
  }
  else {
    my $f = $tables->{$table}->{f};
    print "$table\t$f\n";
    if ($f =~ /^$cfn_dir/) {
      my($c, $b) = readcmap($f);
      %{$$map} = %$c;
      $tables->{$table}->{mmap} = {%$b};
      printf("%d\t{%d}\n", %$c + 0, %$b + 0) if ($debug);
    }
    elsif ($f =~ /^$uni_dir/) {
      my($u) = readumap($f);
      %{$$map} = %$u;
      printf("%d\n", %$u + 0) if ($debug);
      #printf("%d\t{%d}\n", %$u + 0, %$v + 0) if ($debug);
    }
    else {
      next;
    }
  }
  if ($table =~ /^ms-cp/ || ($table =~ /UCS2$/ && $table !~ /^Adobe/)) {
    $tables->{"i-$table"}->{map} = {};
    my $imap = \$tables->{"i-$table"}->{map};
    foreach my $c (keys %{$$map}) {
      $$imap->{$$map->{$c}} = $c;
    }
  }
}

my $ibridges = { # since 20010330, based on the idea of Hideyuki Suzuki
  'Adobe-CNS1,Unicode'		=> {
    'CID'		=> [ 'Adobe-CNS1-UCS2' ],
  },
  'Adobe-GB1,Unicode'		=> {
    'CID'		=> [ 'Adobe-GB1-UCS2' ],
  },
  'Adobe-Japan1,Unicode'	=> {
    'CID'		=> [ 'Adobe-Japan1-UCS2' ],
  },
  'Adobe-Korea1,Unicode'	=> {
    'CID'		=> [ 'Adobe-Korea1-UCS2' ],
  },
};
my $cbridges = {
  'Adobe-CNS1,Big5'		=> {
    'UniCNS-UCS2'	=> [ 'i-ETen-B5-UCS2' ],
  },
  'Adobe-CNS1,Unicode'		=> {
    'ETen-B5'		=> [ 'ETen-B5-UCS2' ],
    'B5pc'		=> [ 'B5pc-UCS2' ],
  },
  'Adobe-GB1,PRC'		=> {
    'UniGB-UCS2'	=> [ 'i-GBK-EUC-UCS2' ],
  },
  'Adobe-GB1,Unicode'		=> {
    'GBK-EUC'		=> [ 'GBK-EUC-UCS2' ],
    'GBKp-EUC'		=> [ 'GBK-EUC-UCS2' ],
    'GBpc-EUC'		=> [ 'GBpc-EUC-UCS2' ],
  },
  'Adobe-Japan1,ShiftJIS'	=> {
    'UniJIS-UCS2'	=> [ 'i-90ms-RKSJ-UCS2' ],
  },
  'Adobe-Japan1,Unicode'	=> {
    '90ms-RKSJ'		=> [ '90ms-RKSJ-UCS2' ],
    '90msp-RKSJ'	=> [ '90ms-RKSJ-UCS2' ],
    '90pv-RKSJ'		=> [ '90pv-RKSJ-UCS2' ],
  },
  'Adobe-Korea1,Wansung'	=> {
    'UniKS-UCS2'	=> [ 'i-KSCms-UHC-UCS2' ],
  },
  'Adobe-Korea1,Unicode'	=> {
    'KSCms-UHC'		=> [ 'KSCms-UHC-UCS2' ],
    'KSCpc-EUC'		=> [ 'KSCpc-EUC-UCS2' ],
  },
};
my $ubridges = { # has no relation to PostScript version
  'Adobe-CNS1,Unicode'		=> {
    'ETen-B5'		=> [ 'ms-cp950' ],
    'B5pc'		=> [ 'mac-big5' ],
  },
  'Adobe-GB1,Unicode'		=> {
    'GBK-EUC'		=> [ 'ms-cp936' ],
    'GBKp-EUC'		=> [ 'ms-cp936' ],
    'GBpc-EUC'		=> [ 'mac-gb' ],
    #'GBTpc-EUC'	=> [ 'mac-gbt' ],	# same effect above
  },
  'Adobe-Japan1,Unicode'	=> {
    '90ms-RKSJ'		=> [ 'ms-cp932' ],
    '90msp-RKSJ'	=> [ 'ms-cp932' ],
    '90pv-RKSJ'		=> [ 'mac-japanese' ],
  },
  'Adobe-Korea1,Unicode'	=> {
    'KSCms-UHC'		=> [ 'ms-cp949' ],
    'KSCpc-EUC'		=> [ 'mac-korean' ],
    #'KSC-EUC'		=> [ 'euc-kr' ],	# since 20001207
    'KSC-EUC'		=> [ 'ascii' ],		# same effect above
    #'KSC-Johab'	=> [ 'johab' ],		# no effect
  },
  #
  'Adobe-CNS1,Big5'		=> {
    'UniCNS-UCS2'	=> [ 'i-ms-cp950' ],	# since 20001203
  },
  'Adobe-GB1,PRC'		=> {
    'UniGB-UCS2'	=> [ 'i-ms-cp936' ],	# since 20001203
  },
  'Adobe-Japan1,ShiftJIS'	=> {
    'UniJIS-UCS2'	=> [ 'i-ms-cp932' ],	# since 20001203
  },
  'Adobe-Korea1,Wansung'	=> {
    'UniKS-UCS2'	=> [ 'i-ms-cp949' ],	# since 20001203
    'KSC-EUC'		=> [ 'ascii' ],		# since 20001207
  },
  #'Adobe-Korea1,Johab'		=> {
    #'UniKS-UCS2'	=> [ 'i-johab' ],	# no effect
  #},
};
my $icbridges = {
  'Adobe-CNS1,Big5'		=> {
    'UniCNS-UCS2'	=> [ 'i-ETen-B5-UCS2' ],
  },
  'Adobe-CNS1,Unicode'		=> {
    'CID'		=> [ 'Adobe-CNS1-UCS2' ],
    'ETen-B5'		=> [ 'ETen-B5-UCS2' ],
    'B5pc'		=> [ 'B5pc-UCS2' ],
  },
  'Adobe-GB1,PRC'		=> {
    'UniGB-UCS2'	=> [ 'i-GBK-EUC-UCS2' ],
  },
  'Adobe-GB1,Unicode'		=> {
    'CID'		=> [ 'Adobe-GB1-UCS2' ],
    'GBK-EUC'		=> [ 'GBK-EUC-UCS2' ],
    'GBKp-EUC'		=> [ 'GBK-EUC-UCS2' ],
    'GBpc-EUC'		=> [ 'GBpc-EUC-UCS2' ],
  },
  'Adobe-Japan1,ShiftJIS'	=> {
    'UniJIS-UCS2'	=> [ 'i-90ms-RKSJ-UCS2' ],
  },
  'Adobe-Japan1,Unicode'	=> {
    'CID'		=> [ 'Adobe-Japan1-UCS2' ],
    '90ms-RKSJ'		=> [ '90ms-RKSJ-UCS2' ], # no effect
    '90msp-RKSJ'	=> [ '90ms-RKSJ-UCS2' ], # no effect
    '90pv-RKSJ'		=> [ '90pv-RKSJ-UCS2' ], # no effect
  },
  'Adobe-Korea1,Wansung'	=> {
    'UniKS-UCS2'	=> [ 'i-KSCms-UHC-UCS2' ],
  },
  'Adobe-Korea1,Unicode'	=> {
    'CID'		=> [ 'Adobe-Korea1-UCS2' ],
    'KSCms-UHC'		=> [ 'KSCms-UHC-UCS2' ],
    'KSCpc-EUC'		=> [ 'KSCpc-EUC-UCS2' ],
  },
};
my $icubridges = { # has no relation to PostScript version
  'Adobe-CNS1,Big5'		=> {
    'UniCNS-UCS2'	=> [ 'i-ETen-B5-UCS2',		'i-ms-cp950' ],
  },
  'Adobe-CNS1,Unicode'		=> {
    'CID'		=> [ 'Adobe-CNS1-UCS2' ],
    'ETen-B5'		=> [ 'ETen-B5-UCS2',		'ms-cp950' ],
    'B5pc'		=> [ 'B5pc-UCS2',		'mac-big5' ],
  },
  'Adobe-GB1,PRC'		=> {
    'UniGB-UCS2'       => [ 'i-GBK-EUC-UCS2',		'i-ms-cp936' ],
  },
  'Adobe-GB1,Unicode'		=> {
    'CID'		=> [ 'Adobe-GB1-UCS2' ],
    'GBK-EUC'		=> [ 'GBK-EUC-UCS2',		'ms-cp936' ],
    'GBKp-EUC'		=> [ 'GBK-EUC-UCS2',		'ms-cp936' ],
    'GBpc-EUC'		=> [ 'GBpc-EUC-UCS2',		'mac-gb' ],
  },
  'Adobe-Japan1,ShiftJIS'	=> {
    'UniJIS-UCS2'	=> [ 'i-90ms-RKSJ-UCS2',	'i-ms-cp932' ],
  },
  'Adobe-Japan1,Unicode'	=> {
    'CID'		=> [ 'Adobe-Japan1-UCS2' ],
    '90ms-RKSJ'		=> [ '90ms-RKSJ-UCS2',		'ms-cp932' ],
    '90msp-RKSJ'	=> [ '90ms-RKSJ-UCS2',		'ms-cp932' ],
    '90pv-RKSJ'		=> [ '90pv-RKSJ-UCS2',		'mac-japanese' ],
  },
  'Adobe-Korea1,Wansung'	=> {
    'UniKS-UCS2'	=> [ 'i-KSCms-UHC-UCS2',	'i-ms-cp949' ],
    'KSC-EUC'		=> [ 'ascii' ],
  },
  'Adobe-Korea1,Unicode'	=> {
    'CID'		=> [ 'Adobe-Korea1-UCS2' ],
    'KSCms-UHC'		=> [ 'KSCms-UHC-UCS2',		'ms-cp949' ],
    'KSCpc-EUC'		=> [ 'KSCpc-EUC-UCS2',		'mac-korean' ],
    'KSC-EUC'		=> [ 'ascii' ],
  },
};
my $bridges = undef;
$bridges = undef	if ($GEN eq '0.81');
$bridges = $ubridges	if ($GEN eq '0.92');
$bridges = $ibridges	if ($GEN eq '0.925');
$bridges = $cbridges	if ($GEN eq '0.93');
$bridges = $icbridges	if ($GEN eq '0.935' || !$GEN);
$bridges = $icubridges	if ($GEN eq '0.94');

sub bridge { # bridge gaps of CID space
  my($reg, $ord, $code, $id, $fi, @f) = @_;
  my(@h, @v) = ((), ());
  my(@g) = () if ($DEBUG2);
  my($c_b, $n_b) = (0, scalar(keys(%{$bridges->{"$reg-$ord,$code"}})));
  #for (my $i=1; $i<@f; $i++) {
  for (my $i=0; $i<@f; $i++) { # since 20010330
    last if ($c_b >= $n_b && !$DEBUG2); # since 20010402, for efficiency
    if ($i != $fi) {
      next if ($f[$i] eq '*');
      my $name = $cmap_name{"$reg,$ord"}->[$i];
      if (!$bridges->{"$reg-$ord,$code"}->{$name}) {
	push(@g, {'name'=>$name, 'code'=>$f[$i]}) if ($DEBUG2);
	next;
      }
      $c_b++; # since 20010402, for efficiency
      foreach $bridge_name (@{$bridges->{"$reg-$ord,$code"}->{$name}}) {
	next if (!$tables->{$bridge_name}->{map});
	my $map = \$tables->{$bridge_name}->{map};
	my @c = split(/,/, $f[$i]);
	foreach my $c (@c) {
	  my $b = 0;
	  if ($c =~ /(.*)v$/) {
	    $c = hex($1);
	    if (!($b = $$map->{$c})) {
	      ;
	    }
	    elsif ($code =~ /Unicode/ && 0xfffd == $b) {
	      # need for ToUnicode CMap
	      ;
	    }
	    elsif ($code =~ /Unicode/ && (0xf860 <= $b && $b <= 0xf868)) {
	      # need for Unicode.org table
	      ;
	    }
	    else {
	      if (!grep($_ == $b, @v)) { # push uniquely
		push(@v, $b);
		printf DEBUG2 ("%s-%s-%s-V\t%d\t%0*x\t%0*x\t%s\n",
			       $ord, $code,
			       ($name!~/^CID/)?$name:"Identity",
			       $id, ($c<=0xff&&$name!~/^(Uni|CID)/)?2:4,
			       $c, ($b<=0xff)?2:4, $b,
			       $bridge_name) if ($DEBUG2);
	      }
	    }
	  }
	  else {
	    $c = hex($c) if ($i != 0);
	    if (!($b = $$map->{$c})) {
	      ;
	    }
	    elsif ($code =~ /Unicode/ && 0xfffd == $b) {
	      # need for ToUnicode CMap
	      ;
	    }
	    elsif ($code =~ /Unicode/ && (0xf860 <= $b && $b <= 0xf868)) {
	      # need for Unicode.org table
	      ;
	    }
	    else {
	      if (!grep($_ == $b, @h)) { # push uniquely
		push(@h, $b);
		printf DEBUG2 ("%s-%s-%s-H\t%d\t%0*x\t%0*x\t%s\n",
			       $ord, $code,
			       ($name!~/^CID/)?$name:"Identity",
			       $id, ($c<=0xff&&$name!~/^(Uni|CID)/)?2:4,
			       $c, ($b<=0xff)?2:4, $b,
			       $bridge_name) if ($DEBUG2);
	      }
	    }
	  }
	}
	(!@h && !@v) && push(@g, {'name'=>$name, 'code'=>$f[$i], 'bridge'=>$bridge_name}) if ($DEBUG2);
      }
    }
    last if ((@h || @v) && $rapid_bridge); # since 20010402, '-rapid_bridge'
  }
  if ($DEBUG2) {
    if (!@h && !@v) {
      foreach my $g (@g) {
	foreach my $c (split(/,/, $g->{code})) {
	  if ($g->{name} =~ /(Roman|Katakana|Hiragana|Hankaku)/) {
	    $c = hex($c);
	    printf DEBUG2 ("#%s-%s.%s\t%d\t%0*x\t%s\n",
			   $ord, $code, $g->{name},
			   $id, ($c<=0xff&&$g->{name}!~/^(Uni|CID)/)?2:4,
			   $c, $g->{bridge});
	  }
	  else {
	    if ($c =~ /(.*)v$/) {
	      $c = hex($1);
	      printf DEBUG2 ("#%s-%s-%s-V\t%d\t%0*x\t%s\n",
			     $ord, $code,
			     ($g->{name}!~/^CID/)?$g->{name}:"Identity",
			     $id, ($c<=0xff&&$g->{name}!~/^(Uni|CID)/)?2:4,
			     $c, $g->{bridge});
	    }
	    else {
	      $c = hex($c) if ($g->{name}!~/^CID/);
	      printf DEBUG2 ("#%s-%s-%s-H\t%d\t%0*x\t%s\n",
			     $ord, $code,
			     ($g->{name}!~/^CID/)?$g->{name}:"Identity",
			     $id, ($c<=0xff&&$g->{name}!~/^(Uni|CID)/)?2:4,
			     $c, $g->{bridge});
	    }
	  }
	}
	last;
      }
    }
  }
  @h = sort numerically @h;
  @v = sort numerically @v;
  return(\@h, \@v);
}

open(DEBUG2, ">icmap+tops.pl.log") || die "$!" if ($DEBUG2);
for my $imap (sort keys %$cid2ttf) {
  my($reg, $ord, $sup) = ($imap =~ /(.*),(.*),.*/, $cid2ttf->{$imap}->{sup});
  my($reg_ord, $code) = ("$reg-$ord", $imap =~ /.*,.*,(.*)/);
  my $base = $cid2ttf->{$imap}->{base};
  open(IN, "$cid2code->{$reg_ord}") || die "$!";
  open(DEBUG, ">$cid2code->{$reg_ord}_$base") || die "$!" if ($DEBUG);
  print("$out_dir/$reg_ord-$code\n");
  print DEBUG2 ("$out_dir/$reg_ord-$code\n") if ($DEBUG2);
  open(OUT, ">$out_dir/$reg_ord-$code") || die "$!";
  my($fi) = undef;
  my($n, %cm, %dcm, %vcm, %vdcm) = (0, (), (), (), ());
  my($n_id, $n_cm, $n_dcm, $n_vcm, $n_vdcm) = (0, 0, 0, 0, 0) if ($debug);
  my($n_bm, $n_vbm) = (0, 0) if ($debug);
  while (<IN>) {
    chomp;
    if (/^#/) {
      next;
    }
    elsif (!$fi && /^CID(\t.*)+/) {
      my @f = split(/\t+/, $_);
      for (my $i=0; $i<@f; $i++) {
	$fi = $i, last if ($f[$i] eq $base);
      }
      ($fi) || die "error: not found $base";
      printf DEBUG ("CID\t%s\n", $f[$fi]) if ($DEBUG);
      push(@{$cmap_name{"$reg,$ord"}}, @f);
    }
    elsif ($fi && /^\d+\t/) {
      my @f = split(/\t+/, $_);
      my $id = $f[0] + 0;
      $n_id = $id if ($debug);
      my(@h, @v) = ((), ());
      ($f[$fi] eq '*') && printf DEBUG ("%d\t%s\n", $id, '*') if ($DEBUG);
      if ($f[$fi] eq '*') {
	my($h, $v) = bridge($reg, $ord, $code, $id, $fi, @f);
	@h = @$h, @v = @$v;
	next if (!@h && !@v);
	$n_bm++ if ($debug && @h);
	$n_vbm++ if ($debug && @v);
      }
      else {
	my @c = split(/,/, $f[$fi]);
	#printf DEBUG ("%d\t%s\n", $id, $f[$fi]) if ($DEBUG);
	foreach my $c (@c) {
	  if ($c =~ /(.*)v$/) { 
	    push(@v, hex($1));
	  }
	  else {
	    push(@h, hex($c));
	  }
	}
	@h = sort numerically @h; # doubly sure
	@v = sort numerically @v; # doubly sure
      }
      $cm{$id} = $h[0] if (@h > 0);
      $dcm{$id} = [@h] if (@h > 1);
      $vcm{$id} = $v[0] if (@v > 0);
      $vdcm{$id} = [@v] if (@v > 1);
      if ($code =~ /Unicode/) {
	printf DEBUG ("%d\t%s\n", $id,
		      join(',', map(sprintf("%0*x", 4, $_),@h),
			   map(sprintf("%0*xv", 4, $_),@v))) if ($DEBUG);
      }
      else {
	printf DEBUG ("%d\t%s\n", $id,
		      join(',', map(sprintf("%0*x", ($_<=0xff)?2:4, $_),@h),
			   map(sprintf("%0*xv", ($_<=0xff)?2:4, $_),@v))) if ($DEBUG);
      }
      $n = $id;
    }
  }
  close(DEBUG) if ($DEBUG);
  close(IN);
  my $id;
  print OUT<<EOF;
%!PS-Adobe-3.0 Resource-TrueType
%%Creator: $creator
%%BeginResource: TrueType ($reg_ord-$code)
7 dict begin
/Registry ($reg) def
/Ordering ($ord) def
/Supplement $sup def
/Code [
EOF
  #
  # Code
  #
  for ($id = 0; $id <= $n; $id++) {
    my $c = $cm{$id};
    if ($cm{$id} == 0) { 
      # mix vertical glyphs
      $c = $vcm{$id};
    }
    elsif ($code =~ /Unicode/ && $vcm{$id} != 0 &&
	   # mix vertical glyphs in R (Restrict) region
	   # except for full-width latin and half-width katakana
	   0xf900 <= $cm{$id} && !(0xff00 <= $cm{$id} && $cm{$id} <= 0xff9f)) {
      $c = $vcm{$id};
    }
    printf OUT ("%s16#%04x%s",
		(($id % 8 == 0 && $id) ? "\n" : ($id ? ' ': '')),
		$c, ($id == $n) ? "\n" : "");
    $n_cm += ($c!=0)?1:0 if ($debug);
  }
  #
  print OUT<<EOF;
] def
/DuplicateCodeMap [
EOF
  #
  # DuplicateCodeMap
  #
  for $id (sort numerically keys %dcm) {
    printf OUT ("[ %d\t", $id);
    foreach $c (sort numerically @{$dcm{$id}}) {
      printf OUT ("16#%04x ", $c);
    }
    printf OUT ("]\n");
    $n_dcm++ if ($debug);
  }
  #
  print OUT<<EOF;
] def
/VerticalCodeMap [
EOF
  #
  # VerticalCodeMap
  #
  for $id (sort numerically keys %vcm) {
    if ($cm{$id} == 0) {
      # mix vertical glyphs
      ;
      $n_vcm++ if ($debug);
    }
    elsif ($code =~ /Unicode/ && $vcm{$id} != 0 &&
	   # mix vertical glyphs in R (Restrict) region
	   # except for full-width latin and half-width katakana
	   0xf900 <= $cm{$id} && !(0xff00 <= $cm{$id} && $cm{$id} <= 0xff9f)) {
      ;
      $n_vcm++ if ($debug);
    }
    else {
      print OUT '%'; # do not mix vertical glyphs, otherwise
    }
    printf OUT ("[ %d\t16#%04x ]\n", $id, $vcm{$id});
  }
  #
  print OUT<<EOF;
] def
/VerticalDuplicateCodeMap [
EOF
  #
  # VerticalDuplicateCodeMap
  #
  for $id (sort numerically keys %vdcm) {
    printf OUT ("[ %d\t", $id);
    foreach $c (sort numerically @{$vdcm{$id}}) {
      printf OUT ("16#%04x ", $c);
    }
    printf OUT ("]\n");
    $n_vdcm++ if ($debug);
  }
  #
  print OUT<<EOF;
] def
currentdict end /$reg_ord-$code exch /TrueType defineresource pop
%%EndResource
%%EOF
EOF
  close(OUT);
  printf("%d\t(+%d\t%d)\t%d\t(+%d\t%d)\t%d\n", $n_cm, $n_bm, $n_dcm, $n_vcm, $n_vbm, $n_vdcm, $n_id) if ($debug);
}
close(DEBUG2) if ($DEBUG2);

sub numerically {
  $a <=> $b;
}

sub readcmap {
  my($f) = @_;
  my($dir, $IN) = ($f =~ m|^(.*)/([^/]+)$|);
  open($IN, "$macdos2 $f|") || die "$!";
  my($in_cr, $in_bc, $in_br) = (0, 0, 0);
  my(%c, %b) = ((), ());
  while (<$IN>) {
    my $flag = 0;
    chomp;
    if ($_ =~ /\/([^\s]+)\s+usecmap/) {
      my($c, $b) = readcmap("$dir/$1");
      %c = %$c, %b = %$b;
    }
    my($sl, $su, $d, @d) = (undef, undef, undef, ());
    $in_cr = !0, next if ($_ =~ /[^\s]+\s+begincidrange/);
    $in_cr = 0, next if ($_ =~ /endcidrange/);
    $in_br = !0, next if ($_ =~ /[^\s]+\s+beginbfrange/);
    $in_br = 0, next if ($_ =~ /endbfrange/);
    $in_bc = !0, next if ($_ =~ /[^\s]+\s+beginbfchar/);
    $in_bc = 0, next if ($_ =~ /endbfchar/);
    if ($in_cr) {
      if ($_ =~ /<([0-9a-fA-F]+)>\s+<([0-9a-fA-F]+)>\s+([0-9]+)/) {
	($sl, $su, $d) = (hex($1), hex($2), $3);
	$flag = !0;
      }
    }
    if ($in_br) {
      if ($_ =~ /<([0-9a-fA-F]+)>\s+<([0-9a-fA-F]+)>\s+(<[0-9a-fA-F]+>)/) {
	($sl, $su, $d) = (hex($1), hex($2), $3);
	$flag = !0;
      }
    }
    if ($in_bc) {
      if ($_ =~ /<([0-9a-fA-F]+)>\s+(<[0-9a-fA-F]+>)/) {
	($sl, $su, $d) = (hex($1), hex($1), $2);
	$flag = !0;
      }
    }
    if ($flag) {
      if ($d =~ /<([0-9a-fA-F]+)>/) {
	$d = $1;
	$d =~ s/([0-9a-fA-F]{4})/$1,/g;
	@d = map(hex($_), split(/,/, $d));
      }
      else {
	push(@d, $d + 0);
      }
      #printf("%04X %s\n", $sl, join('',map(sprintf("%04X",$_),@d)));
      #printf("%04X %04X %s\n", $sl, $su, join('',map(sprintf("%d",$_),@d)));
      for (my $s = $sl; $s <= $su; $s++) {
	if (@d > 1) {
	  $b{$s} = [@d[0..@d-2], $d[@d-1] + ($s-$sl)];
	}
	else {
	  $c{$s} = $d[@d-1] + ($s-$sl);
	}
      }
    }
  }
  close($IN);
  return(\%c, \%b);
}

sub readumap {
  my($f) = @_;
  if (-f "$f.gz") {
    open(IN, "zcat $f.gz|$macdos2|") || die "$!";
  }
  else {
    open(IN, "$macdos2 $f|") || die "$!";
  }
  my(%u) = ();
  while (<IN>) {
    chomp;
    next if (/^#/);
    my @f = split(/\t+/, $_);
    my $c = hex($f[0]);
    my @u = map(hex($_), split(/\+/, $f[1]));
    my $u = $u[0]; # skimping!
    next if ($c == 0 || $u == 0);
    #printf("0x%04x\t0x%04x\n", $c, $u);
    $u{$c} = $u;
  }
  close(IN);
  #
  $u{0xa8bf} = 0xe7c8 if ($f eq "$uni_dir/VENDORS/APPLE/CHINSIMP.TXT");
  #
  return(\%u);
}
